# invoke watcher

import re
from typing import Generator

from invoke.watchers import StreamWatcher

from .common import parse_int


class LineWatcher(StreamWatcher):
    def __init__(self, *, file_callback=None, progress_callback=None, line_callback=None):
        self.index = 0
        self.file_pattern = re.compile(r"^~([^\s]+)\s([0-9,]+)\s(.*)")
        self.percent_pattern = re.compile(r"^\s*([^\s]+)\s+([0-9\.]+)%")
        self.file_callback = file_callback
        self.progress_callback = progress_callback
        self.line_callback = line_callback
        self.file_count = 0
        self.byte_count = 0
        super().__init__()

    def parse_line(self, line):
        if self.line_callback:
            return self.line_callback(line)
        file = self.file_pattern.match(line)
        if file:
            codes, length, filename = file.groups()
            self.file_count += 1
            return self.file_callback(filename, self.file_count, parse_int(length), codes)
        percent = self.percent_pattern.match(line)
        if percent:
            length, progress = percent.groups()
            length = parse_int(length)
            chunk = length - self.byte_count
            self.byte_count = length
            return self.progress_callback(chunk, progress)

    def submit(self, stream: str) -> Generator[str, None, None]:
        while True:
            try:
                pos = stream.index("\n", self.index)
            except ValueError:
                try:
                    pos = stream.index("\r", self.index)
                except ValueError:
                    return
            chunk = stream[self.index : pos]  # noqa: E203
            self.index = pos + 1
            chunks = chunk.split("\n")
            for chunk in chunks:
                for line in chunk.split("\r"):
                    if len(line):
                        self.parse_line(line)
        yield ""



"""Console script for cptree."""

import sys

import click
import click.core

from .cptree import cptree
from .exception_handler import ExceptionHandler
from .shell import _shell_completion
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"

DEFAULT_HASH = "sha256"
HASHES = ["md5", "sha1", "sha256", "sha512"]


def _ehandler(ctx, option, debug):
    ctx.obj = dict(ehandler=ExceptionHandler(debug))
    ctx.obj["debug"] = debug


FLAG_CHOICES = ["ask", "force", "never"]
HASH_CHOICES = list(HASHES) + ["none"]
PROGRESS_CHOICES = ["enable", "ascii", "none"]


@click.command("cptree", context_settings={"auto_envvar_prefix": "CPTREE"})
@click.version_option(message=header)
@click.option(
    "-d",
    "--debug",
    is_eager=True,
    is_flag=True,
    callback=_ehandler,
    help="debug mode",
)
@click.option(
    "--shell-completion",
    is_flag=False,
    flag_value="[auto]",
    callback=_shell_completion,
    help="configure shell completion",
)
@click.option(
    "-c",
    "--create",
    type=click.Choice(FLAG_CHOICES),
    default="ask",
    help="create destination directory if nonexistent",
)
@click.option(
    "-D",
    "--delete",
    type=click.Choice(FLAG_CHOICES + ["force-no-countdown"]),
    default="never",
    help="delete destination directory before transfer",
)
@click.option(
    "-p",
    "--progress",
    type=click.Choice(PROGRESS_CHOICES),
    default="enable",
    help="total transfer progress",
)
@click.option(
    "-o",
    "--output_dir",
    type=click.Path(file_okay=False, writable=True),
    help="output directory",
)
@click.option(
    "-h",
    "--hash",
    type=click.Choice(HASH_CHOICES),
    default=DEFAULT_HASH,
    help="select checksum hash",
)
@click.option(
    "-r/-R",
    "--rsync/--no-rsync",
    is_flag=True,
    default=True,
    help="enable/disable rsync transfer",
)
@click.option("-a", "--rsync-args", help="rsync pass-through arguments")
@click.option("-f", "--file-list", is_flag=True, help="scan only, output file list")
@click.argument("SRC")
@click.argument("DST")
@click.pass_context
def cli(
    ctx,
    debug,
    shell_completion,
    create,
    delete,
    progress,
    output_dir,
    file_list,
    hash,
    rsync,
    rsync_args,
    src,
    dst,
):
    """rsync transfer with progress indicator and checksum verification"""
    if create == "never":
        create = False
    if delete == "never":
        delete = False
    if hash == "none":
        hash = None
    if progress == "none":
        progress = None

    if src[-1] != "/":
        if file_list:
            click.echo(f"WARNING: changing source to {src + '/'}", err=True)
            src += "/"
        else:
            if click.confirm(f"Change source to {src + '/'} ?", default="Y"):
                src += "/"

    return cptree(
        src,
        dst,
        create=create,
        delete=delete,
        progress=progress,
        output_dir=output_dir,
        hash=hash,
        rsync=rsync,
        rsync_args=rsync_args,
        file_list=file_list,
    )


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



# generate checksum for a list of files

import atexit
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory

import click
from invoke import run
from tqdm import tqdm

from .common import host_mode, runner, split_target, which
from .exceptions import ChecksumCompareFailed, ChecksumGenerationFailed
from .watcher import LineWatcher


def is_remote(host):
    return bool(host)


def is_local(host):
    return not bool(host)


def checksum(target, hash, output_file, tqdm_kwargs=None, src=None, dst=None):
    """generate BSD-style checksum for each file in target, returning local file containing result"""

    if tqdm_kwargs is None:
        tqdm_kwargs = dict(disable=True)

    host, base = split_target(target)
    base = Path(base.rstrip("/"))

    if is_local(host):
        base = base.resolve()

    if src:
        label = "source"
    elif dst:
        label = "destination"
    else:
        raise RuntimeError

    click.echo(f"Generating checksums for {host_mode(host)} {label} {target}")

    # try linux command without breaking
    hash_cmd = which(hash + "sum", host, quiet=True)
    if hash_cmd:
        # add option if linux
        hash_cmd += " --tag"
    else:
        # try bsd-style hash command
        hash_cmd = which(hash, host)

    with NamedTemporaryFile("w+", delete=False) as tempfile:

        with tqdm(unit=" lines", **tqdm_kwargs) as bar:

            atexit.register(delete_file, tempfile.name)

            def _line(line):
                bar.update(1)

            cmd = f"cd {str(base)}; {which('find', host)} . -type f -exec {hash_cmd} \\{{\\}} \\;"
            genproc = runner(host)(
                cmd,
                warn=True,
                watchers=[LineWatcher(line_callback=_line)],
                hide=True,
                in_stream=False,
                out_stream=tempfile.file,
            )

            if genproc.failed:
                raise ChecksumGenerationFailed(genproc.stderr)

        tempfile.close()

        # sort checksums into output file
        with output_file.open("w") as ofp:
            run(f"{which('sort')} {tempfile.name}", in_stream=False, out_stream=ofp, hide=True)

    return output_file


def delete_file(filename):
    Path(filename).unlink()


def compare_checksums(src_sums, dst_sums):
    """run diff on hash digest files, return length if identical, otherwise raise exception"""

    diff = run(f"{which('diff')} {str(src_sums)} {str(dst_sums)}", in_stream=False, hide=True, warn=True)
    if diff.failed:
        output_dir = Path(src_sums).parent
        (output_dir / "cptree.diff.out").write_text(diff.stdout)
        (output_dir / "cptree.diff.err").write_text(diff.stderr)
        tempdir = TemporaryDirectory(prefix="cptree", delete=False)
        shutil.copytree(output_dir, tempdir)
        raise ChecksumCompareFailed("details written to {tempdir.name}")
    wc = run(f"{which('wc')} -l {str(src_sums)}", in_stream=False, hide=True)
    return int(wc.stdout.split()[0])



# global test config

import os
from pathlib import Path

import pytest


@pytest.fixture
def remote_host():
    return os.environ["TEST_REMOTE_HOST"]


@pytest.fixture
def remote_user():
    return os.environ["TEST_REMOTE_USER"]


@pytest.fixture
def remote_dir():
    return os.environ["TEST_REMOTE_DIR"]


@pytest.fixture
def test_datadir():
    return Path(os.environ["TEST_DATA_DIR"])


@pytest.fixture
def local_src(test_datadir):
    return str(test_datadir / "src") + "/"


@pytest.fixture
def local_dst(test_datadir):
    return str(test_datadir / "dst")


@pytest.fixture
def remote_src(remote_user, remote_host, remote_dir):
    return f"{remote_user}@{remote_host}:{remote_dir}/src/"


@pytest.fixture
def remote_dst(remote_user, remote_host, remote_dir):
    return f"{remote_user}@{remote_host}:{remote_dir}/dst"


@pytest.fixture
def big_local_src():
    dir = str(Path(os.environ["TEST_BIG_SRC"]).resolve())
    if dir[-1] != "/":
        dir += "/"
    return dir



import io
import sys
from pathlib import Path

from tqdm import tqdm


class ProgressReader:
    def __init__(self, in_stream, **kwargs):
        self.tqdm = None
        self.buffer = ""
        kwargs.setdefault("miniters", 1)

        if in_stream.seekable():
            head = in_stream.tell()
            tail = in_stream.seek(0, io.SEEK_END)
            in_stream.seek(head, io.SEEK_SET)
            kwargs.setdefault("total", tail - head)
            kwargs.setdefault("unit", " bytes")
            kwargs.setdefault("unit_scale", True)
            self.count = 0
        else:
            kwargs.setdefault("unit", " lines")
            self.count = None
        self.kwargs = kwargs
        self.in_stream = in_stream

    def __enter__(self):
        self.tqdm = tqdm(**self.kwargs)
        return self

    def __exit__(self, *args):
        if self.tqdm is not None:
            self.tqdm.close()
            self.tqdm = None
        return False

    def _update(self, len):
        if self.tqdm.total:
            self.tqdm.update(len + 1)
        else:
            self.tqdm.update(1)

    def readlines(self):
        while True:
            chunk = self.in_stream.readline()
            self.buffer += chunk
            while "\n" in self.buffer:
                line, _, self.buffer = self.buffer.partition("\n")
                self._update(len(line))
                yield line
            if not chunk:
                if self.buffer:
                    self._update(len(self.buffer))
                    yield self.buffer
                return


if __name__ == "__main__":
    import time

    import click

    @click.command("progress")
    @click.option("-d", "--delay", type=float)
    @click.argument("INFILE", type=click.Path(dir_okay=False, readable=True), default="-")
    @click.argument("OUTFILE", type=click.Path(dir_okay=False, writable=True), default="-")
    def progress(delay, infile, outfile):
        """progress test cli"""

        closers = []

        if infile in ["-", None]:
            infile = sys.stdin
        else:
            infile = Path(infile).open("r")
            closers.insert(0, infile)
        if infile in ["-", None]:
            outfile = sys.stdout
        else:
            outfile = Path(outfile).open("w")
            closers.insert(0, outfile)

        try:
            with ProgressReader(infile) as reader:
                for line in reader.lines():
                    outfile.write(line + "\n")
                    if delay:
                        time.sleep(delay)
        finally:
            for closer in closers:
                closer.close()

    sys.exit(progress())



# utility functions

from pathlib import Path

import fabric
import invoke

from .exceptions import CommandNotFound


def split_target(target: str) -> (str, str):
    """return (hostname, path) for target; hostname is None for local target"""
    host, _, path = str(target).partition(":")
    if path:
        return host, path
    else:
        return None, host


def parse_int(field):
    return int(field.replace(",", ""))


def read_file_lines(filename, strip=False):
    with Path(filename).open("r") as ifp:
        for line in ifp.readline():
            if strip:
                line = line.strip()
            if line:
                yield line


def write_file_lines(dir, name, lines):
    dir = Path(dir)
    if not dir.is_dir():
        dir.mkdir()
    file = dir / name
    if isinstance(lines, (list, tuple)):
        lines = "\n".join(lines)
    if lines:
        lines = lines.rstrip("\n") + "\n"
    file.write_text(lines)


def runner(host):
    return fabric.Connection(host).run if host else invoke.run


def host_mode(host):
    return "remote" if host else "local"


def which(command, host=None, quiet=False):
    """return local or remote command path if valid, otherwise raise exception or optionally return None"""
    probe = runner(host)(f"which 2>/dev/null {command}", in_stream=False, hide=True, warn=True)
    cmd = probe.stdout.strip()
    if probe.ok and cmd:
        return cmd
    elif quiet:
        return None
    else:
        raise CommandNotFound(f"{command} not available on {host_mode(host)}")



# directory tests

import pytest
from invoke import run

from cptree import cptree
from cptree.exceptions import InvalidDirectory, RsyncTransferFailed


@pytest.fixture(autouse=True)
def delete_local_dst(local_dst):
    print()
    cmd = f"rm -rf {local_dst}"
    run(cmd)


# def test_cp_dev():
#    ret = cptree("/dev", "/tmp/dev", force_create=True)
#    assert ret == 0


def test_cp_local_local(local_src, local_dst):
    ret = cptree(local_src, local_dst, create=True, delete="force-no-countdown")
    assert ret == 0


def test_cp_local_remote(local_src, remote_dst):
    ret = cptree(local_src, remote_dst, create=True, delete="force-no-countdown")
    assert ret == 0


def test_cp_remote_local(remote_src, local_dst):
    ret = cptree(remote_src, local_dst, create=True, delete="force-no-countdown")
    assert ret == 0


def test_cp_badlocal_badlocal():
    with pytest.raises(InvalidDirectory):
        cptree("bad_local_src", "/bad_local_dst", create=None, delete=None)


def test_cp_local_badlocal(local_src):
    with pytest.raises(RsyncTransferFailed):
        cptree(local_src, "/bad_local_dst", create=None, delete=None)


def test_cp_badlocal_local(local_dst):
    with pytest.raises(InvalidDirectory):
        cptree("bad_local_src", local_dst, create=True, delete="force-no-countdown")


def test_cp_badremote_local(remote_host, local_dst):
    bad_remote_src = f"{remote_host}:bad_remote_src"
    with pytest.raises(InvalidDirectory):
        cptree(bad_remote_src, local_dst, create=True, delete="force-no-countdown")


def test_cp_local_badremote(local_src, remote_host):
    bad_remote_dst = f"{remote_host}:/bad_remote_dst"
    with pytest.raises(RsyncTransferFailed):
        cptree(local_src, bad_remote_dst, create=False, delete=False)


def test_cp_progress(big_local_src, remote_dst, test_datadir):
    cptree(big_local_src, remote_dst, create=None, delete="force", hash="sha256", output_dir=test_datadir / "output")



# verify directory existence

import os
from pathlib import Path
from time import sleep

import click
from fabric import Connection
from invoke import run

from .common import split_target
from .exceptions import InvalidDirectory

DELETE_COUNTDOWN = 10


def _verify_directory(host, target, dir_type, create=None, delete=None):
    target = str(target)
    if host:
        label = f"Remote {dir_type} {host}:{target}"
        runner = Connection(host).run
        if dir_type == "output":
            raise InvalidDirectory("Unsupported: " + label)
        if "~" in str(target):
            raise InvalidDirectory("Illegal '~' expansion: " + label)
    else:
        if "~" in str(target):
            target = os.path.expanduser(str(target))
            target = str(Path(target).resolve())
        label = f"Local {dir_type} {target}"
        runner = run

    result = runner(f"[ -d {target} ]", warn=True, hide=True)
    if result.failed:
        # doesn't exist, check for create
        if (dir_type in ["destination", "output"]) and (create in ["ask", "force", True]):
            if create == "ask":
                click.confirm(f"{label} does not exist. Create it?", abort=True)
            runner(f"mkdir -p {target}", echo=True)
        elif dir_type != "destination":
            # not created, all but DST must exist
            raise InvalidDirectory(label)
    else:
        # exists, check for delete request
        if (dir_type in ["destination", "output"]) and delete in [
            "ask",
            "force",
            "force-no-countdown",
            True,
        ]:
            cmd = confirm_delete(target, label, host, delete)
            runner(cmd, echo=True)


def confirm_delete(target, label, host, delete):
    cmd = f"rm -rf {target}"
    if delete == "ask":
        click.confirm(f"Delete {label}?", abort=True)
    if delete != "force-no-countdown":
        delete_countdown(label, cmd, host)
    return cmd


def delete_countdown(msg, cmd, host):
    click.echo(f"\nCAUTION: About to delete {msg}")
    click.echo(f"\nExecuting '{cmd}' on {host or 'localhost'} in ", nl=False)
    for count in range(DELETE_COUNTDOWN, 0, -1):
        click.echo(f"{count}...", nl=False)
        sleep(1)


def verify_output_directory(target):
    host, target = split_target(target)
    return _verify_directory(host, target, "output", "ask", None)


def verify_src_directory(target):
    host, target = split_target(target)
    return _verify_directory(host, target, "source", None, None)


def verify_dst_directory(target, create=None, delete=None):
    if delete:
        create = None
    host, target = split_target(target)
    # verify parent of DST dir exists
    _verify_directory(host, Path(target).parent, "destination", create, False)
    if delete:
        # delete target dir
        _verify_directory(host, target, "destination", None, delete)



# checksum test cases

import subprocess

import pytest
from invoke import run

from cptree.checksum import checksum
from cptree.cptree import cptree

HASH = "sha256"


@pytest.fixture
def output_dir(test_datadir):
    return test_datadir / "output"


@pytest.fixture
def count(local_src):
    return int(subprocess.check_output(f"cd {local_src.rstrip('/')}; find . -type f | wc -l", shell=True, text=True))


@pytest.fixture
def kwargs(count):
    return dict(total=count, disable=False, ascii=True, ncols=120, miniters=1)


@pytest.fixture
def compare_checksums(output_dir, count):
    def _compare_checksums(test_sums):
        reference_sums = (output_dir / "reference").with_suffix(test_sums.suffix)
        cmd = f"diff -y {str(reference_sums)} {str(test_sums)}"
        result = run(cmd, hide=True, warn=True)
        if result.failed:
            assert False, dict(stdout=result.stdout.split("\n"), stderr=result.stderr.split("\n"))
        return True

    return _compare_checksums


def test_checksum_local_src(local_src, local_dst, output_dir, compare_checksums, kwargs):
    test_sums = checksum(local_src, HASH, output_dir / f"local.src.{HASH}", kwargs, src=True)
    assert compare_checksums(test_sums)


def test_checksum_local_dst(local_src, local_dst, output_dir, compare_checksums, kwargs):
    assert cptree(local_src, local_dst, delete="force-no-countdown", hash=None) == 0
    test_sums = checksum(local_dst, HASH, output_dir / f"local.dst.{HASH}", kwargs, dst=True)
    assert compare_checksums(test_sums)


def test_checksum_remote_src(remote_src, local_src, local_dst, output_dir, compare_checksums, kwargs):
    assert cptree(local_src, remote_src, delete="force-no-countdown", hash=None) == 0
    test_sums = checksum(remote_src, HASH, output_dir / f"remote.src.{HASH}", kwargs, src=True)
    assert compare_checksums(test_sums)


def test_checksum_remote_dst(local_src, remote_dst, output_dir, compare_checksums, kwargs):
    assert cptree(local_src, remote_dst, delete="force-no-countdown", hash=None) == 0
    test_sums = checksum(remote_dst + "/", HASH, output_dir / f"remote.dst.{HASH}", kwargs, dst=True)
    assert compare_checksums(test_sums)



"""Top-level package for cptree."""

from .cli import cli
from .cptree import cptree
from .version import __author__, __email__, __timestamp__, __version__

__all__ = [
    "cli",
    "cptree",
    "__version__",
    "__timestamp__",
    "__author__",
    "__email__",
]



# cptree exceptions


class Fail(Exception):
    pass


class InvalidDirectory(Fail):
    pass


class ChecksumCompareFailed(Fail):
    pass


class RsyncTransferFailed(Fail):
    pass


class ChecksumGenerationFailed(Fail):
    pass


class CommandNotFound(Fail):
    pass


class UnsupportedRsyncArgument(Fail):
    pass


class UnrecognizedRsyncPrescanOutput(Fail):
    pass



# cptree implementation

import re
import shlex
import shutil
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

import click
from invoke import run
from tqdm import tqdm

from .checksum import checksum, compare_checksums
from .common import parse_int, which, write_file_lines
from .exceptions import (
    RsyncTransferFailed,
    UnrecognizedRsyncPrescanOutput,
    UnsupportedRsyncArgument,
)
from .verify import verify_dst_directory, verify_output_directory, verify_src_directory
from .watcher import LineWatcher

NAME_LENGTH = 12

RESERVED_RSYNC_ARGS = [
    "-P",
    "--progress",
    "--info",
    "-i",
    "--itemize-changes",
    "--out-format" "--debug",
    "--msgs2stderr",
    "-q",
    "--quiet",
    "-v",
    "--verbose",
]


def mkopts(kwargs):
    opts = " "
    for k, v in kwargs:
        if v:
            opts.append(f"{k} {v} ")
        else:
            opts.append(f"{k} ")
    opts = opts.strip()
    return opts


def prescan(src, dst, cmd, opts, file_list):
    if not file_list:
        click.echo("Scanning...\r", nl=False)
    scanproc = run(f"{cmd} -a --list-only {opts} {src} {dst}", in_stream=False, hide=True, warn=True)
    if scanproc.ok:
        items = scanproc.stdout.strip().split("\n")
    else:
        click.echo(scanproc.stderr, err=True)
        sys.exit(scanproc.return_code)
    counts = {
        "d": 0,
        "-": 0,
        "l": 0,
        "p": 0,
        "c": 0,
        "b": 0,
    }
    names = {
        "d": "directories",
        "-": "files",
        "l": "links",
        "p": "pipes",
        "c": "devices",
        "b": "devices",
    }

    pattern = re.compile(r"^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s(.*)$")
    size = 0
    files = []
    for item in items:
        match = pattern.match(item)
        if not match:
            raise UnrecognizedRsyncPrescanOutput(item)
        fields = match.groups()
        file_length = parse_int(fields[1])
        size += int(file_length)
        file_type = item[0]
        counts[file_type] += 1
        if file_type == "-":
            files.append(fields[4])

    details = []
    for k, v in counts.items():
        if v:
            details.append(f"{names[k]}={v}")

    if (counts["b"] + counts["c"]) > 0:
        if file_list:
            click.output('WARNING: Transfer includes devices.', err=True) 
        else:
            click.confirm(
                "Transfer includes devices.  Are you certain you want to proceed?",
                abort=True,
            )

    if not file_list:
        click.echo(f"Transferring {len(items)} items: [{', '.join(details)}]")

    return size, items, files


def cptree(*args, output_dir=None, **kwargs):
    """call _cptree with work_dir from argument or a temp dir"""

    if output_dir:
        output_dir = Path(output_dir)
        if not output_dir.is_dir():
            output_dir.mkdir()
        kwargs["output_dir"] = output_dir
        return _cptree(*args, **kwargs)
    else:
        with TemporaryDirectory() as temp_dir:
            kwargs["output_dir"] = Path(temp_dir)
            return _cptree(*args, **kwargs)


def _verify_dirs(src, dst, output_dir, create, delete):
    verify_output_directory(output_dir)
    verify_src_directory(src)
    verify_dst_directory(dst, create, delete)


def _rsync_echo(cmd):
    """return rsync command line without our added progress formatting options"""
    hide_options = [("--info", True), ("--out-format", True), ("--progress", False)]
    words = shlex.split(cmd)
    for word, arg in hide_options:
        pos = words.index(word)
        words.pop(pos)
        if arg:
            words.pop(pos)
    click.echo(shlex.join(words))


def _check_rsync_args(args):
    """ensure progress formatting args not present in user rsync args"""
    if args is None:
        return ""
    args = shlex.split(args)
    for reserved in RESERVED_RSYNC_ARGS:
        if reserved in args:
            raise UnsupportedRsyncArgument(f"{repr(reserved)} not supported in --rsync-args")
    return shlex.join(args)


def _cptree(  # noqa: C901
    src,
    dst,
    *,
    create=None,
    delete=None,
    progress=True,
    output_dir=None,
    hash=None,
    rsync=True,
    rsync_args=None,
    file_list=False,
):
    _verify_dirs(src, dst, output_dir, create, delete)

    rsync_cmd = which("rsync")
    rsync_args = _check_rsync_args(rsync_args)

    total_bytes, items, files = prescan(src, dst, rsync_cmd, rsync_args, file_list)
    total_items = len(items)

    if file_list:
        pattern = re.compile(r'\S+\s+\S+\s+\S+\s+\S+\s+(\S+)')
        for item in items:
            match = pattern.match(item)
            filename = match.groups()[0]
            click.echo(filename)
        return 0

    file_list = write_file_lines(output_dir, "cptree.files", ["./" + f for f in files])

    tqdm_kwargs = dict(
        total=total_bytes,
        disable=bool(progress in ["disable", False, None]),
        ascii=bool(progress == "ascii"),
        ncols=shutil.get_terminal_size().columns - 1,
        miniters=1,
        delay=1,
    )

    bar = tqdm(unit="B", unit_scale=True, **tqdm_kwargs)

    def _line(line):
        click.echo(line)

    def _file(filename, item_count, length, codes):
        count = str(item_count).zfill(len(str(total_items)))
        bar.set_description(f"[{count}/{total_items}]", refresh=False)

    def _progress(bytes_read, percent):
        bar.update(bytes_read)

    if progress:
        progress_args = "--progress --info PROGRESS2 --out-format '~%i %l %f'"
        watcher = LineWatcher(file_callback=_file, progress_callback=_progress)
    else:
        progress_args = ""
        watcher = LineWatcher(line_callback=_line)

    if rsync:

        cmd = f"{rsync_cmd} -avz {rsync_args} {progress_args} {src} {dst}"
        _rsync_echo(cmd)

        proc = run(
            cmd,
            watchers=[watcher],
            in_stream=False,
            hide=True,
            warn=True,
        )
        rsync_stderr = proc.stderr.strip().split("\n")
    else:
        click.echo("Skipping rsync transfer")
        rsync_stderr = []
        proc = None

    bar.close()

    for error in rsync_stderr:
        if error.strip():
            click.echo(error, err=True)

    if proc is not None and proc.failed:
        raise RsyncTransferFailed(f"rsync failed with error code: {proc.return_code}")

    tqdm_kwargs["total"] = len(files)
    if hash:
        _verify_hashes(src, dst, hash, output_dir, tqdm_kwargs)

    if proc is None:
        return 0
    else:
        return proc.return_code


def _verify_hashes(src, dst, hash, output_dir, tqdm_kwargs):
    src_sums = checksum(src, hash, output_dir / f"cptree.src.{hash}", tqdm_kwargs, src=True)
    dst_sums = checksum(dst, hash, output_dir / f"cptree.dst.{hash}", tqdm_kwargs, dst=True)
    count = compare_checksums(src_sums, dst_sums)
    click.echo(f"\nSuccessful Transfer. Verified matching {hash.upper()} hashes on {count} files.\n")



# hash type tests

import os
import re
from pathlib import Path

import pytest
from fabric import Connection
from invoke import run


@pytest.fixture
def hashes():
    result = run("cptree --help")
    match = re.search(r".*\s--hash\s+\[(.*)\]", result.stdout)
    hlist = match.groups()[0]
    ret = hlist.split("|")
    ret.remove("none")
    return ret


@pytest.fixture
def hosts():
    return os.environ["TEST_HOSTS"].split(",")


def test_hash_hosts(hosts, hashes, local_src):
    local_src = str(Path(local_src).resolve()) + "/"
    for host in hosts:
        assert Connection(host).run("uname -a").ok
        for hash in hashes:
            assert run(
                f"cptree {local_src} {host}:archive/cptree_test --create=force --hash {hash} --output /tmp/test/output"
            ).ok
